@@ -14,142 +14,171 @@ the line and cell magics, respectively.
 Usage
 =====
 
-To enable the magics below, execute ``%load_ext rpy2.ipython``.
+To enable the magics below, execute `%load_ext rpy2.ipython`.
 
-``%R``
+`%R`
 
 {R_DOC}
 
-``%Rpush``
+`%Rpush`
 
 {RPUSH_DOC}
 
-``%Rpull``
+`%Rpull`
 
 {RPULL_DOC}
 
-``%Rget``
+`%Rget`
 
 {RGET_DOC}
 
 """
 
-#-----------------------------------------------------------------------------
+# -----------------------------------------------------------------------------
 #  Copyright (C) 2012 The IPython Development Team
-#  Copyright (C) 2013-2017 rpy2 authors
+#  Copyright (C) 2013-2019 rpy2 authors
 #
 #  Distributed under the terms of the BSD License.  The full license is in
 #  the file COPYING, distributed as part of this software.
-#-----------------------------------------------------------------------------
+# -----------------------------------------------------------------------------
 
+import contextlib
 import sys
 import tempfile
 from glob import glob
 from os import stat
 from shutil import rmtree
+import textwrap
 
 # numpy and rpy2 imports
 
 import rpy2.rinterface as ri
+import rpy2.rinterface_lib.callbacks
 import rpy2.robjects as ro
 import rpy2.robjects.packages as rpacks
+from rpy2.robjects.lib import grdevices
 from rpy2.robjects.conversion import (Converter,
-                                      localconverter)
+                                      localconverter,
+                                      get_conversion)
 import warnings
-from rpy2.robjects.conversion import converter as template_converter
 
 # Try loading pandas and numpy, emitting a warning if either cannot be
 # loaded.
 try:
-    import numpy
+    import numpy  # type: ignore
     try:
-        import pandas
-    except ImportError:
+        import pandas  # type: ignore
+    except ImportError as ie:
         pandas = None
-        warnings.warn(' '.join(("The Python package 'pandas' is strongly"
-                                "recommended when using `rpy2.ipython`.",
-                                "Unfortunately it could not be loaded,",
-                                "but at least we found 'numpy'.")))
-except ImportError:
+        warnings.warn('The Python package `pandas` is strongly '
+                      'recommended when using `rpy2.ipython`. '
+                      'Unfortunately it could not be loaded '
+                      '(error: %s), '
+                      'but at least we found `numpy`.' % str(ie))
+except ImportError as ie:
     # Give up on numerics
     numpy = None
-    warnings.warn(' '.join(("The Python package 'pandas' is strongly",
-                            "recommended when using `rpy2.ipython`.",
-                            "Unfortunately it could not be loaded,",
-                            "as we did not manage to load 'numpy'",
-                            "first.")))
-if numpy:
-  from rpy2.robjects import numpy2ri
-  template_converter += numpy2ri.converter
-  if pandas:
-    from rpy2.robjects import pandas2ri
-    template_converter += pandas2ri.converter
+    pandas = None
+    warnings.warn('The Python package `pandas` is strongly '
+                  'recommended when using `rpy2.ipython`. '
+                  'Unfortunately it could not be loaded, '
+                  'as we did not manage to load `numpy` '
+                  'in the first place (error: %s).' % str(ie))
 
+# IPython imports.
 
-# IPython imports
-
-import IPython
-from IPython.core.displaypub import publish_display_data
-from IPython.core.magic import (cell_magic,
-                                Magics,
+from IPython.core import displaypub  # type: ignore
+from IPython.core.magic import (Magics,   # type: ignore
                                 magics_class,
                                 line_cell_magic,
                                 line_magic,
-                                needs_local_scope)
-from IPython.core.magic_arguments import (argument,
+                                needs_local_scope,
+                                no_var_expand)
+from IPython.core.magic_arguments import (argument,  # type: ignore
                                           argument_group,
                                           magic_arguments,
                                           parse_argstring)
 
-# TODO: drop support for ipython < 5.0
-try:
-    from IPython.external.simplegeneric import generic
-except ImportError:
-    # IPython 4.0
-    from simplegeneric import generic
+template_converter = get_conversion()
 
 
-class RInterpreterError(ri.RRuntimeError):
+def _get_ipython_template_converter(template_converter=template_converter,
+                                    numpy=numpy, pandas=pandas):
+    if numpy:
+        from rpy2.robjects import numpy2ri
+        template_converter += numpy2ri.converter
+        if pandas:
+            from rpy2.robjects import pandas2ri
+            template_converter += pandas2ri.converter
+    return template_converter
+
+
+def _get_converter(template_converter=template_converter):
+    return Converter('ipython conversion',
+                     template=template_converter)
+
+
+ipy_template_converter = _get_ipython_template_converter(template_converter,
+                                                         numpy=numpy,
+                                                         pandas=pandas)
+converter = _get_converter(template_converter=ipy_template_converter)
+
+
+def CELL_DISPLAY_DEFAULT(res, args):
+    return ro.r.show(res)
+
+
+class RInterpreterError(ri.embedded.RRuntimeError):
     """An error when running R code in a %%R magic cell."""
-    
+
+    msg_prefix_template = ('Failed to parse and evaluate line %r.\n'
+                           'R error message: %r')
+    rstdout_prefix = '\nR stdout:\n'
+
     def __init__(self, line, err, stdout):
         self.line = line
         self.err = err.rstrip()
         self.stdout = stdout.rstrip()
 
     def __str__(self):
-        s = 'Failed to parse and evaluate line %r.\nR error message: %r' % \
-                (self.line, self.err)
+        s = (self.msg_prefix_template %
+             (self.line, self.err))
         if self.stdout and (self.stdout != self.err):
-            s += '\nR stdout:\n' + self.stdout
+            s += self.rstdout_prefix + self.stdout
         return s
 
 
-converter = Converter('ipython conversion',
-                      template=template_converter)
-
-
-# The default conversion for lists is currently to make them an R list. That has
-# some advantages, but can be inconvenient (and, it's inconsistent with the way
-# python lists are automatically converted by numpy functions), so for
-# interactive use in the rmagic, we call unlist, which converts lists to vectors
-# **if the list was of uniform (atomic) type**.
-@converter.ri2py.register(list)
-def ri2py_list(obj):
+# The default conversion for lists is currently to make them an R list. That
+# has some advantages, but can be inconvenient (and, it's inconsistent with
+# the way python lists are automatically converted by numpy functions), so
+# for interactive use in the rmagic, we call unlist, which converts lists to
+# vectors **if the list was of uniform (atomic) type**.
+@converter.py2rpy.register(list)
+def py2rpy_list(obj):
     # simplify2array is a utility function, but nice for us
     # TODO: use an early binding of the R function
-    return ro.r.simplify2array(obj)
+    cv = ro.conversion.get_conversion()
+    robj = ri.ListSexpVector(
+            [cv.py2rpy(x) for x in obj]
+        )
+    res = ro.r.simplify2array(robj)
+    # The current default converter for the ipython rmagic
+    # might make `res` a numpy array. We need to ensure that
+    # a rpy2 objects is returned (issue #866).
+    res_rpy = cv.py2rpy(res)
+    return res_rpy
 
 
-# The R magic is opiniated about what the R vectors should become.
-@converter.ri2ro.register(ri.SexpVector)
-def _(obj):
-    if 'data.frame' in obj.rclass:
-        # request to turn it to a pandas DataFrame
-        res = converter.ri2py(obj)
-    else:
-        res = ro.sexpvector_to_ro(obj)
-    return res        
+# TODO: remove ?
+# # The R magic is opiniated about what the R vectors should become.
+# @converter.ri2ro.register(ri.SexpVector)
+# def _(obj):
+#     if 'data.frame' in obj.rclass:
+#         # request to turn it to a pandas DataFrame
+#         res = converter.rpy2py(obj)
+#     else:
+#         res = ro.sexpvector_to_ro(obj)
+#     return res
 
 
 @magics_class
@@ -169,11 +198,11 @@ class RMagics(Magics):
                     the magic's current converter is used.
 
         cache_display_data : bool
-            If True, the published results of the final call to R are 
+            If True, the published results of the final call to R are
             cached in the variable 'display_cache'.
 
         device : ['png', 'X11', 'svg']
-            Device to be used for plotting. 
+            Device to be used for plotting.
             Currently only 'png', 'X11' and 'svg' are supported,
             with 'png' and 'svg' being most useful in the notebook,
             and 'X11' allowing interactive plots in the terminal.
@@ -181,48 +210,45 @@ class RMagics(Magics):
         """
         super(RMagics, self).__init__(shell)
         self.cache_display_data = cache_display_data
-
         self.Rstdout_cache = []
-
         self.converter = converter
-        
         self.set_R_plotting_device(device)
 
     def set_R_plotting_device(self, device):
         """
         Set which device R should use to produce plots.
-        If device == 'svg' then the package 'Cairo' 
+        If device == 'svg' then the package 'Cairo'
         must be installed. Because Cairo forces "onefile=TRUE",
         it is not posible to include multiple plots per cell.
 
-        Parameters
-        ----------
-
-        device : ['png', 'X11', 'svg']
-            Device to be used for plotting. 
-            Currently only "png" and "X11" are supported,
-            with 'png' and 'svg' being most useful in the notebook,
-            and 'X11' allowing interactive plots in the terminal.
+        :param device: ['png', 'X11', 'svg']
+          Device to be used for plotting.
+          Currently only "png" and "X11" are supported,
+          with 'png' and 'svg' being most useful in the notebook,
+          and 'X11' allowing interactive plots in the terminal.
 
         """
         device = device.strip()
         if device not in ['png', 'X11', 'svg']:
-            raise ValueError("device must be one of ['png', 'X11' 'svg'], got '%s'", device)
+            raise ValueError(
+                "device must be one of ['png', 'X11' 'svg'], got '%s'", device)
         if device == 'svg':
             try:
                 self.cairo = rpacks.importr('Cairo')
-            except ri.RRuntimeError as rre:
+            except ri.embedded.RRuntimeError as rre:
                 if rpacks.isinstalled('Cairo'):
-                    msg = "An error occurred when trying to load the R package Cairo'\n%s" % str(rre)
+                    msg = ('An error occurred when trying to load the ' +
+                           'R package Cairo\'\n%s' % str(rre))
                 else:
-                    msg = """
-The R package 'Cairo' is required but it does not appear to be installed/available. Try:
+                    msg = textwrap.dedent("""
+                    The R package 'Cairo' is required but it does not appear
+                    to be installed/available. Try:
 
-import rpy2.robjects.packages as rpacks
-utils = rpacks.importr('utils')
-utils.chooseCRANmirror(ind=1)
-utils.install_packages('Cairo')
-"""
+                    import rpy2.robjects.packages as rpacks
+                    utils = rpacks.importr('utils')
+                    utils.chooseCRANmirror(ind=1)
+                    utils.install_packages('Cairo')
+                    """)
                 raise RInterpreterError(msg)
         self.device = device
 
@@ -241,23 +267,29 @@ utils.install_packages('Cairo')
         boolean indicating whether the return value would be
         visible if the line of code were evaluated in an R REPL.
 
-        R Code evaluation and visibility determination are done via an R call of
-        the form withVisible(code_string), and this entire expression needs to
-        be evaluated in R (we can't use rpy2 function proxies here, as
+        R Code evaluation and visibility determination are done via an R call
+        of the form withVisible(code_string), and this entire expression needs
+        to be evaluated in R (we can't use rpy2 function proxies here, as
         withVisible is a LISPy R function).
-
         """
-        old_writeconsole_regular = ri.get_writeconsole_regular()
-        ri.set_writeconsole_regular(self.write_console_regular)
-        try:
-            # Need the newline in case the last line in code is a comment
-            value, visible = ro.r("withVisible({%s\n})" % code)
-        except (ri.RRuntimeError, ValueError) as exception:
-            warning_or_other_msg = self.flush() # otherwise next return seems to have copy of error
-            raise RInterpreterError(code, str(exception), warning_or_other_msg)
-        text_output = self.flush()
-        ri.set_writeconsole_regular(old_writeconsole_regular)
-        return text_output, value, visible[0]
+        with contextlib.ExitStack() as stack:
+            if self.cache_display_data:
+                stack.enter(
+                    rpy2.rinterface_lib
+                    .callbacks.obj_in_module(rpy2.rinterface_lib.callbacks,
+                                             'consolewrite_print',
+                                             self.write_console_regular)
+                )
+            try:
+                # Need the newline in case the last line in code is a comment.
+                value, visible = ro.r("withVisible({%s\n})" % code)
+            except (ri.embedded.RRuntimeError, ValueError) as exception:
+                # Otherwise next return seems to have copy of error.
+                warning_or_other_msg = self.flush()
+                raise RInterpreterError(code, str(exception),
+                                        warning_or_other_msg)
+            text_output = self.flush()
+            return text_output, value, visible[0]
 
     def write_console_regular(self, output):
         """
@@ -312,7 +344,7 @@ utils.install_packages('Cairo')
                     # variable
                     raise NameError("name '%s' is not defined" % input)
 
-            with localconverter(self.converter) as cv:
+            with localconverter(self.converter):
                 ro.r.assign(input, val)
 
     # @skip_doctest
@@ -344,21 +376,21 @@ utils.install_packages('Cairo')
 
 
         This is useful when a structured array is desired as output, or
-        when the object in R has mixed data types. 
+        when the object in R has mixed data types.
         See the %%R docstring for more examples.
 
         Notes
         -----
 
-        Beware that R names can have '.' so this is not fool proof.
-        To avoid this, don't name your R objects with '.'s...
+        Beware that R names can have dots ('.') so this is not fool proof.
+        To avoid this, don't name your R objects with dots...
 
         """
         args = parse_argstring(self.Rpull, line)
         outputs = args.outputs
-        with localconverter(self.converter) as cv:
+        with localconverter(self.converter):
             for output in outputs:
-                robj = ri.globalenv.get(output)
+                robj = ri.globalenv.find(output)
                 self.shell.push({output: robj})
 
     # @skip_doctest
@@ -371,13 +403,16 @@ utils.install_packages('Cairo')
     @line_magic
     def Rget(self, line):
         """
-        Return an object from rpy2, possibly as a structured array (if possible).
-        Similar to Rpull except only one argument is accepted and the value is 
+        Return an object from rpy2, possibly as a structured array (if
+        possible).
+        Similar to Rpull except only one argument is accepted and the value is
         returned rather than pushed to self.shell.user_ns::
 
             In [3]: dtype=[('x', '<i4'), ('y', '<f8'), ('z', '|S1')]
 
-            In [4]: datapy = np.array([(1, 2.9, 'a'), (2, 3.5, 'b'), (3, 2.1, 'c'), (4, 5, 'e')], dtype=dtype)
+            In [4]: datapy = np.array([(1, 2.9, 'a'), (2, 3.5, 'b'),
+            ...                        (3, 2.1, 'c'), (4, 5, 'e')],
+            ...                        dtype=dtype)
 
             In [5]: %R -i datapy
 
@@ -392,21 +427,20 @@ utils.install_packages('Cairo')
         output = args.output
         # get the R object with the given name, starting from globalenv
         # in the search path
-        with localconverter(self.converter) as cv:
-            res = ro.globalenv.get(output[0])
+        with localconverter(self.converter):
+            res = ro.globalenv.find(output[0])
         return res
 
-
     def setup_graphics(self, args):
         """Setup graphics in preparation for evaluating R code.
 
         args : argparse bunch (should be whatever the R magic got)."""
-        
+
         if getattr(args, 'units') is not None:
             if args.units != "px" and getattr(args, 'res') is None:
                 args.res = 72
 
-        plot_arg_names = ['width', 'height', 'pointsize', 'bg']
+        plot_arg_names = ['width', 'height', 'pointsize', 'bg', 'type']
         if self.device == 'png':
             plot_arg_names += ['units', 'res']
 
@@ -419,22 +453,22 @@ utils.install_packages('Cairo')
         tmpd = None
         if self.device in ['png', 'svg']:
             # Create a temporary directory for R graphics output
-            # TODO: Do we want to capture file output for other device types other
-            # than svg & png?
+            # TODO: Do we want to capture file output for other device types
+            # other than svg & png?
             tmpd = tempfile.mkdtemp()
             tmpd_fix_slashes = tmpd.replace('\\', '/')
 
             if self.device == 'png':
                 # Note: that %% is to pass into R for interpolation there
-                ro.r.png("%s/Rplots%%03d.png" % tmpd_fix_slashes,
-                        **argdict)
+                grdevices.png("%s/Rplots%%03d.png" % tmpd_fix_slashes,
+                              **argdict)
             elif self.device == 'svg':
                 self.cairo.CairoSVG("%s/Rplot.svg" % tmpd_fix_slashes,
                                     **argdict)
 
         elif self.device == 'X11':
-            # Open a new X11 device, except if the current one is already an X11
-            # device
+            # Open a new X11 device, except if the current one is already an
+            # X11 device.
             ro.r("""
             if (substr(names(dev.cur()), 1, 3) != "X11") {
                 X11()
@@ -442,7 +476,8 @@ utils.install_packages('Cairo')
 
         else:
             # TODO: This isn't actually an R interpreter error...
-            raise RInterpreterError("device must be one of ['png', 'X11' 'svg']")
+            raise RInterpreterError(
+                'device must be one of ("png", "X11", "svg")')
 
         return tmpd
 
@@ -453,16 +488,16 @@ utils.install_packages('Cairo')
             Probably provided by some tmpdir call
         isolate_svgs : bool
             Enable SVG namespace isolation in metadata"""
-        
+
         # read in all the saved image files
         images = []
         display_data = []
 
-        # Default empty metadata dictionary
+        # Default empty metadata dictionary.
         md = {}
 
         if self.device == 'png':
-            for imgfile in sorted( glob("%s/Rplots*png" % graph_dir) ):
+            for imgfile in sorted(glob('%s/Rplots*png' % graph_dir)):
                 if stat(imgfile).st_size >= 1000:
                     with open(imgfile, 'rb') as fh_img:
                         images.append(fh_img.read())
@@ -473,37 +508,43 @@ utils.install_packages('Cairo')
             # -- empty ones are not published
             if stat(imgfile).st_size >= 1000:
                 with open(imgfile, 'rb') as fh_img:
-                    images.append(fh_img.read())
+                    images.append(fh_img.read().decode())
 
-        mimetypes = { 'png' : 'image/png', 'svg' : 'image/svg+xml' }
+        mimetypes = {'png': 'image/png', 'svg': 'image/svg+xml'}
         mime = mimetypes[self.device]
 
-        # By default, isolate SVG images in the Notebook to avoid garbling
+        # By default, isolate SVG images in the Notebook to avoid garbling.
         if images and self.device == "svg" and isolate_svgs:
             md = {'image/svg+xml': dict(isolated=True)}
 
-        # flush text streams before sending figures, helps a little with output
+        # Flush text streams before sending figures, helps a little with
+        # output.
         for image in images:
-            # synchronization in the console (though it's a bandaid, not a real sln)
-            sys.stdout.flush(); sys.stderr.flush()
+            # Synchronization in the console (though it's a bandaid, not a
+            # real solution).
+            sys.stdout.flush()
+            sys.stderr.flush()
             display_data.append(('RMagic.R', {mime: image}))
 
         return display_data, md
 
-
     # @skip_doctest
     @magic_arguments()
     @argument(
         '-i', '--input', action='append',
-        help=('Names of input variable from shell.user_ns to be assigned to R variables'
-              ' of the same names after using the Converter self.converter. Multiple names can be'
-              ' passed separated only by commas with no whitespace.')
+        help=textwrap.dedent("""
+        Names of input variable from `shell.user_ns` to be assigned to R
+        variables of the same names after using the Converter self.converter.
+        Multiple names can be passed separated only by commas with no
+        whitespace.""")
     )
     @argument(
         '-o', '--output', action='append',
-        help=("Names of variables to be pushed from rpy2 to shell.user_ns after executing"
-              " cell body (rpy2's internal facilities will apply ri2ro as appropriate)."
-              " Multiple names can be passed separated only by commas with no whitespace.")
+        help=textwrap.dedent("""
+        Names of variables to be pushed from rpy2 to `shell.user_ns` after
+        executing cell body (rpy2's internal facilities will apply ri2ro as
+        appropriate). Multiple names can be passed separated only by commas
+        with no whitespace.""")
     )
     @argument(
         '-n', '--noreturn',
@@ -531,10 +572,11 @@ utils.install_packages('Cairo')
     @argument_group("SVG", "SVG specific arguments")
     @argument(
         '--noisolation',
-        help=('Disable SVG isolation in the Notebook. By default, SVGs are isolated to'
-              ' avoid namespace collisions between figures.'
-              ' Disabling SVG isolation allows to reference previous figures or share'
-              ' CSS rules across a set of SVGs.'),
+        help=textwrap.dedent("""
+        Disable SVG isolation in the Notebook. By default, SVGs are isolated to
+        avoid namespace collisions between figures. Disabling SVG isolation
+        allows to reference previous figures or share CSS rules across a set
+        of SVGs."""),
         action='store_false',
         default=True,
         dest='isolate_svgs'
@@ -542,29 +584,47 @@ utils.install_packages('Cairo')
     @argument_group("PNG", "PNG specific arguments")
     @argument(
         '-u', '--units', type=str, choices=["px", "in", "cm", "mm"],
-        help=('Units of png plotting device sent as an argument to *png* in R. One of'
-              ' ["px", "in", "cm", "mm"].')
-        )
+        help=textwrap.dedent("""
+        Units of png plotting device sent as an argument to *png* in R. One of
+        ["px", "in", "cm", "mm"]."""))
     @argument(
         '-r', '--res', type=int,
-        help=('Resolution of png plotting device sent as an argument to *png* in R.'
-              ' Defaults to 72 if *units* is one of ["in", "cm", "mm"].')
+        help=textwrap.dedent("""
+        Resolution of png plotting device sent as an argument to *png* in R.
+        Defaults to 72 if *units* is one of ["in", "cm", "mm"].""")
         )
+    @argument(
+        '--type', type=str,
+        choices=['cairo', 'cairo-png', 'Xlib', 'quartz'],
+        help=textwrap.dedent("""
+        Type device used to generate the figure.
+        """))
     @argument(
         '-c', '--converter',
         default=None,
-        help=('Name of local converter to use. A converter contains the rules to convert objects'
-              ' back and forth between Python and R. If not specified/None, the defaut converter'
-              ' for the magic\'s module is used'
-              ' (that is rpy2\'s default converter + numpy converter + pandas converter'
-              ' if all three are available).')
-        )
+        help=textwrap.dedent("""
+        Name of local converter to use. A converter contains the rules to
+        convert objects back and forth between Python and R. If not
+        specified/None, the defaut converter for the magic\'s module is used
+        (that is rpy2\'s default converter + numpy converter + pandas converter
+        if all three are available)."""))
+    @argument(
+        '-d', '--display',
+        default=None,
+        help=textwrap.dedent("""
+        Name of function to use to display the output of an R cell (the last
+        object or function call that does not have a left-hand side
+        assignment). That function will have the signature `(robject, args)`
+        where `robject` is the R objects that is an output of the cell, and
+        `args` a namespace with all parameters passed to the cell.
+        """))
     @argument(
         'code',
         nargs='*',
         )
     @needs_local_scope
     @line_cell_magic
+    @no_var_expand
     def R(self, line, cell=None, local_ns=None):
         """
         Execute code in R, optionally returning results to the Python runtime.
@@ -573,7 +633,8 @@ utils.install_packages('Cairo')
         value to a Python object.  The return value is determined by rpy2's
         behaviour of returning the result of evaluating the final expression.
 
-        Multiple R expressions can be executed by joining them with semicolons::
+        Multiple R expressions can be executed by joining them with
+        semicolons::
 
             In [9]: %R X=c(1,4,5,7); sd(X); mean(X)
             Out[9]: array([ 4.25])
@@ -613,7 +674,8 @@ utils.install_packages('Cairo')
         If cell is not None and line has some R code, it is prepended to
         the R code in cell.
 
-        Objects can be passed back and forth between rpy2 and python via the -i -o flags in line::
+        Objects can be passed back and forth between rpy2 and python via the
+        -i -o flags in line::
 
             In [14]: Z = np.array([1,4,5,10])
 
@@ -672,11 +734,14 @@ utils.install_packages('Cairo')
                 try:
                     converter = self.shell.user_ns[args.converter]
                 except KeyError:
-                    raise NameError("name '%s' is not defined" % args.converter)
+                    raise NameError(
+                        "name '%s' is not defined" % args.converter
+                    )
             if not isinstance(converter, Converter):
-                raise ValueError("'%s' must be a %s object (but it is a %s)."
-                                 % (args.converter, Converter, type(localconverter)))
-            
+                raise TypeError("'%s' must be a %s object (but it is a %s)."
+                                % (args.converter, Converter,
+                                   type(localconverter)))
+
         if args.input:
             for input in ','.join(args.input).split(','):
                 try:
@@ -689,6 +754,17 @@ utils.install_packages('Cairo')
                 with localconverter(converter) as cv:
                     ro.r.assign(input, val)
 
+        if args.display:
+            try:
+                cell_display = local_ns[args.display]
+            except KeyError:
+                try:
+                    cell_display = self.shell.user_ns[args.display]
+                except KeyError:
+                    raise NameError("name '%s' is not defined" % args.display)
+        else:
+            cell_display = CELL_DISPLAY_DEFAULT
+
         tmpd = self.setup_graphics(args)
 
         text_output = ''
@@ -698,49 +774,58 @@ utils.install_packages('Cairo')
                     text_result, result, visible = self.eval(line)
                     text_output += text_result
                 if text_result:
-                    # the last line printed something to the console so we won't return it
+                    # The last line printed something to the console so
+                    # we won't return it.
                     return_output = False
             else:
                 text_result, result, visible = self.eval(code)
                 text_output += text_result
                 if visible:
-                    old_writeconsole_regular = ri.get_writeconsole_regular()
-                    ri.set_writeconsole_regular(self.write_console_regular)
-                    ro.r.show(result)
-                    text_output += self.flush()
-                    ri.set_writeconsole_regular(old_writeconsole_regular)
+                    with contextlib.ExitStack() as stack:
+                        if self.cache_display_data:
+                            stack.enter_context(
+                                rpy2.rinterface_lib
+                                .callbacks
+                                .obj_in_module(rpy2.rinterface_lib
+                                               .callbacks,
+                                               'consolewrite_print',
+                                               self.write_console_regular))
+                        cell_display(result, args)
+                        text_output += self.flush()
 
         except RInterpreterError as e:
             # TODO: Maybe we should make this red or something?
             print(e.stdout)
             if not e.stdout.endswith(e.err):
                 print(e.err)
-            if tmpd:
-                rmtree(tmpd)
-            return
+            raise e
         finally:
             if self.device in ['png', 'svg']:
                 ro.r('dev.off()')
+            if text_output:
+                # display_data.append(('RMagic.R', {'text/plain':text_output}))
+                displaypub.publish_display_data(
+                    data={'text/plain': text_output}, source='RMagic.R')
+            # publish the R images
+            if self.device in ['png', 'svg']:
+                display_data, md = self.publish_graphics(
+                    tmpd, args.isolate_svgs
+                )
 
-        if text_output:
-            # display_data.append(('RMagic.R', {'text/plain':text_output}))
-            publish_display_data(data={'text/plain':text_output}, source='RMagic.R')
-        # publish the R images
-        if self.device in ['png', 'svg']:
-            display_data, md = self.publish_graphics(tmpd, args.isolate_svgs)
-
-            for tag, disp_d in display_data:
-                publish_display_data(data=disp_d, source=tag, metadata=md)
+                for tag, disp_d in display_data:
+                    displaypub.publish_display_data(data=disp_d, source=tag,
+                                                    metadata=md)
 
             # kill the temporary directory - currently created only for "svg"
             # and "png" (else it's None)
-            rmtree(tmpd)
+            if tmpd:
+                rmtree(tmpd)
 
         if args.output:
             with localconverter(converter) as cv:
                 for output in ','.join(args.output).split(','):
-                    output_ipy = ro.globalenv.get(output)
-                    self.shell.push({output: output_ipy })
+                    output_ipy = ro.globalenv.find(output)
+                    self.shell.push({output: output_ipy})
 
         # this will keep a reference to the display_data
         # which might be useful to other objects who happen to use
@@ -749,20 +834,20 @@ utils.install_packages('Cairo')
         if self.cache_display_data:
             self.display_cache = display_data
 
-        # We're in line mode and return_output is still True, 
+        # We're in line mode and return_output is still True,
         # so return the converted result
         if return_output and not args.noreturn:
             if result is not ri.NULL:
                 with localconverter(converter) as cv:
-                    res = cv.ri2py(result)
+                    res = cv.rpy2py(result)
                 return res
 
 
 __doc__ = __doc__.format(
-                R_DOC = ' '*8 + RMagics.R.__doc__,
-                RPUSH_DOC = ' '*8 + RMagics.Rpush.__doc__,
-                RPULL_DOC = ' '*8 + RMagics.Rpull.__doc__,
-                RGET_DOC = ' '*8 + RMagics.Rget.__doc__
+                R_DOC='{0}{1}'.format(' '*8, RMagics.R.__doc__),
+                RPUSH_DOC='{0}{1}'.format(' '*8, RMagics.Rpush.__doc__),
+                RPULL_DOC='{0}{1}'.format(' '*8, RMagics.Rpull.__doc__),
+                RGET_DOC='{0}{1}'.format(' '*8, RMagics.Rget.__doc__)
 )
 
 
@@ -770,15 +855,3 @@ def load_ipython_extension(ip):
     """Load the extension in IPython."""
 
     ip.register_magics(RMagics)
-
-    # TODO: drop support for ipython < 5.0
-    # Initialising rpy2 interferes with readline. Since, at this point, we've
-    # probably just loaded rpy2, we reset the delimiters. See issue gh-2759.
-    if hasattr(ip, 'has_readline'):
-        if ip.has_readline:
-            ip.readline.set_completer_delims(ip.readline_delims)
-    else:
-        # ipython >= 5.0.0
-        # not doing anything (for now)
-        pass
-
